"""

    frosch - Better runtime errors

    Patrick Haller
    betterthannothing.blog
    patrickhaller40@googlemail.com

    License MIT

"""

from .frosch import hook, print_exception

__all__ = ["hook", "print_exception"]



"""

    frosch - Better runtime errors

    Patrick Haller
    betterthannothing.blog
    patrickhaller40@googlemail.com

    License MIT

"""
import importlib
import string
from typing import Optional

class ThemeNotExistsError(Exception):
    """Thrown when trying to import a theme from pygments which does not exist"""

class ConfigManager:
    """Used to parse and store configs from various ways of setting"""

    def __init__(self):
        self.message = None
        self.title = None
        self._theme = None

    def set_notifier(self, title: str, message: str):
        """Setter for notifcation message"""
        self.title = title
        self.message = message

    def has_notifier(self):
        """Returns True if message and/or title is set"""
        if self.message or self.title:
            if not self.message:
                self.message = ""
            if not self.title:
                self.title = ""
            return True
        return False

    def from_kwargs(self, **kwargs):
        """Set all given kwargs as attributes, if attr exists"""
        for key, value in kwargs.items():
            if hasattr(self, key):
                self.__setattr__(key, value)
        return self

    @classmethod
    def default(cls):
        """Constructor with default values"""
        _cls = cls()
        _cls.theme = "monokai"
        return _cls

    @property
    def theme(self):
        """Returns theme object of pygments"""
        return self._theme

    @theme.setter
    def theme(self, theme: str):
        """Sets theme object of pygments for given string"""
        self._theme = ConfigManager._get_theme_from_string(theme)

    @staticmethod
    def _get_theme_from_string(theme_string: str):
        """Import the according theme from pygments"""
        pygment_styles_module = "pygments.styles.{}.{}Style".format(
            theme_string, theme_string.capitalize()
        )

        pygment_styles_module = "pygments.styles.{}".format(
            theme_string
        )

        style_class = "{}Style".format(ConfigManager._capitalize_theme(theme_string))

        try:
            pygment_styles_module = importlib.import_module(pygment_styles_module)
            pygment_style_class = getattr(
                pygment_styles_module,
                style_class
            )
            return pygment_style_class
        except ImportError as import_error:
            raise ThemeNotExistsError("Theme: '{}' does not exists.\nPlease check \
https://github.com/HallerPatrick/frosch#configuration for more infos."
                .format(theme_string)) from import_error
        except AttributeError as attr_error:
            raise ThemeNotExistsError(
                "Could not find style '{}' in module '{}'".format(
                    style_class, pygment_styles_module
                )
            ) from attr_error

    @staticmethod
    def _capitalize_theme(theme_string: str) -> str:
        """Capitalize the theme correctly for import from pygments"""

        # Use this if we only support 3.7 >
        # if theme := ConsoleWriter._get_special_themes(theme_string):
        #    return theme

        theme = ConfigManager._get_special_themes(theme_string)
        if theme:
            return theme

        if "_" in theme_string:
            theme_string = theme_string.replace("_", " ")
        theme_string = string.capwords(theme_string)
        return theme_string.replace(" ", "_")

    @staticmethod
    def _get_special_themes(theme_string: str) -> Optional[str]: # pylint: disable=E1136
        """Some theme names cannot be converted programmaticaly
        (or is just to much time to catch all cases), there
        map theme manually"""

        themes = {
            "bw": "BlackWhite",
            "vs": "VisualStudio",
            "inkpot": "InkPot",
            "paraiso_dark": "ParaisoDark",
            "paraiso_light": "ParaisoLight",
            "rainbow_dash": "RainbowDash",
            "solarized": "SolarizedDark", # Just use dark for now
            "stata_dark": "StataDark",
            "stata_light": "StataLight"
        }

        if theme_string in themes:
            return themes[theme_string]
        return None



"""

    frosch - Better runtime errors

    Patrick Haller
    betterthannothing.blog
    patrickhaller40@googlemail.com

    License MIT

"""

from typing import Any, List

import colorama
from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter
from pygments.lexers.python import Python3Lexer, Python3TracebackLexer

class WrongWriteOrder(Exception):
    """Thrown when the correct order of parts to be written to stream is not correct"""

class Variable:
    """Dataclass for a variable in error throwing line of program"""

    def __init__(self, id_: str, col_offset: int, value: Any = None):
        self.name = id_
        self.col_offset = col_offset
        self.value = value

    def __repr__(self):
        """str representation of a Variable object"""
        return "Variable({!r}, {}, {!r})".format(self.name, self.col_offset, self.value)

    def __eq__(self, other: "Variable"):
        """Mainly for testing purpose"""
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        """We don't care for unique hashes"""
        return hash((self.name, self.col_offset, self.value))

    def tree_str(self):
        """Python>3.8 variable declaration format with types"""
        if self.value is None:
            return "{} = None".format(self.name)
        return "{}: {} = {!r}".format(self.name, type(self.value).__qualname__, self.value)


class ConsoleWriter:
    """Handles formatting, highlighting and writing to output of error message"""

    def __init__(self, theme, stream):
        self.stream = stream
        self.terminal_formater = Terminal256Formatter(style=theme)
        self.python_lexer = Python3Lexer()
        self.python_traceback_lexer = Python3TracebackLexer()
        self.left_offset = None

    def write_traceback(self, traceback_):
        """Highlight traceback and write out"""
        self._write_out(highlight(traceback_, self.python_traceback_lexer, self.terminal_formater))



    @staticmethod
    def offset_vert_lines(offsets) -> str:
        """Construct a line with vertical bars for all offsets in offsets"""
        line = ""

        current_offset = -1

        # First line
        for offset in offsets:
            new_offset = offset - current_offset -1
            line += (" " * new_offset)
            line += "│"
            current_offset = offset

        return line

    @staticmethod
    def left_bar() -> str:
        """Bar used on left side of debug tree"""
        return colorama.Fore.BLUE + "||" + colorama.Style.RESET_ALL

    def _write_out(self, message: str):
        """Write to stream"""
        self.stream.write(message)
        self.stream.flush()

    def write_newline(self):
        """Write newline to stderr"""
        self.stream.write("\n")
        self.stream.flush()


    def write_debug_tree(self, names: List[Variable]):
        """Sort offsets and values for construction of debug tree"""

        sorted_values = list(sorted(names, key=lambda val: val.col_offset))

        num_variables = len(sorted_values)
        try:
            line_offset = " " * self.left_offset
        except TypeError as type_error:
            raise WrongWriteOrder("Offset not defined") from type_error

        lines = [
           "  {}{} ".format(line_offset, self.left_bar()) for _ in range((2*num_variables) + 1)
        ]


        self.construct_debug_tree(lines, sorted_values)

        for line in lines:
            self._write_out(line)
            self.write_newline()

    def construct_debug_tree(self, lines, sorted_values):
        """Construction of debug tree"""
        current_offset = -1
        offsets = [val.col_offset for val in sorted_values]

        unprocessed_values = len(sorted_values)

        # First line
        lines[0] += self.offset_vert_lines(offsets)

        i = 1
        for value in reversed(sorted_values):
            # For every value
            for j in range(unprocessed_values):
                new_offset = sorted_values[j].col_offset - current_offset -1
                lines[i] += (" " * (new_offset))

                if j != (unprocessed_values - 1):
                    lines[i] += "│"
                else:
                    lines[i] += "└── {}".format(
                        highlight(value.tree_str(), self.python_lexer, self.terminal_formater)
                        .rstrip()
                    )

                    # Add spacong row
                    if offsets:
                        del offsets[-1]
                    i += 1
                    lines[i] += self.offset_vert_lines(offsets)

                current_offset = sorted_values[j].col_offset
            unprocessed_values -= 1
            current_offset = -1
            i += 1

        return lines

    def write_last_line(self, lineno: int, line: str):
        """Write out the line which throws runtime error with highlighting"""
        self.write_newline()
        self.left_offset = len(str(lineno))
        self._write_out(" {} {} {}\n".format(
            lineno,
            self.left_bar(),
            highlight(line, self.python_lexer, self.terminal_formater).rstrip())
        )



"""

    frosch - Better runtime errors

    Patrick Haller
    betterthannothing.blog
    patrickhaller40@googlemail.com

    License MIT

"""

import platform
import os

def notify_os(title: str = "Ups", message: str = "Your python program crashed."):
    """Check current OS and run a notification subprocess"""
    current_platform = platform.system()

    if current_platform == "Darwin":
        command = mac_notify(title, message)
    elif current_platform == "Linux":
        command = linux_notify(title, message)
    elif current_platform == "Windows":
        command = windows_notify(title, message)
    else:
        return

    os.system(command)

def mac_notify(title: str, message: str) -> str:
    """Display notification for MacOS systems"""
    command = f'''osascript -e 'display notification "{message}" with title "{title}"' '''
    return command

def linux_notify(title: str, message: str) -> str:
    """Display notification for Linux systems"""
    command = f'''notify-send "{title}" "{message}"'''
    return command

def windows_notify(title: str, message: str) -> str:
    """Display notification for Windows systems"""
    command = f'''powershell -command "$wshell = New-Object -ComObject Wscript.Shell;\
    $wshell.Popup('{message}', 64, '{title}', 0)" '''
    return command



"""

    frosch - Better runtime errors

    Patrick Haller
    betterthannothing.blog
    patrickhaller40@googlemail.com

    License MIT

"""

import ast
import linecache
import sys
import traceback
from bdb import Bdb
import builtins
from contextlib import contextmanager
from collections import ChainMap
from types import TracebackType
from typing import List

from asttokens.util import Token
from colorama import deinit, init
from stack_data import Source
from yapf.yapflib.yapf_api import FormatCode

from .config_manager import ConfigManager
from .notifier import notify_os
from .writer import ConsoleWriter, Variable


class ParseError(Exception):
    """Thrown in crashing line cannot be parsed with ast.parse"""

@contextmanager
def support_windows_colors():
    """Only for windows terminal"""
    init()
    yield
    deinit()

def hook(**kwargs):
    """Initialize configurations for frosch and set hook"""
    if kwargs:
        config_manager = ConfigManager.default().from_kwargs(**kwargs)
    else:
        config_manager = ConfigManager.default()

    pytrace_excepthook.configs = config_manager
    _hook()

def _hook():
    """Overwrite sys.excepthook"""
    # Don't want global vars
    sys.excepthook = pytrace_excepthook

def print_exception(exception: Exception):
    """Pretty print the exception and traceback to stdout"""
    last_stack, line = get_last_stack(exception.__traceback__)
    line = format_line(line)

    variables = get_vars_from_tb(exception.__traceback__, line)

    config_manager = ConfigManager.default()

    # Do we even need exception as param?
    error_type, error_message, traceback_ = sys.exc_info()

    with support_windows_colors():

        console_writer = ConsoleWriter(config_manager.theme, sys.stdout)
        console_writer.write_traceback(
            "".join(traceback.format_exception(
                error_type,
                error_message,
                traceback_
                )
            )
        )
        console_writer.write_last_line(last_stack.lineno, line)
        console_writer.write_debug_tree(variables)
        console_writer.write_newline()


def pytrace_excepthook(error_type: type, error_message: TypeError, traceback_: TracebackType=None):
    """New excepthook to overwrite sys.excepthook"""
    configs = pytrace_excepthook.configs

    last_stack, line = get_last_stack(traceback_)

    line = format_line(line)

    variables = get_vars_from_tb(traceback_, line)

    # Write down
    with support_windows_colors():
        console_writer = ConsoleWriter(configs.theme, sys.stderr)
        console_writer.write_traceback(
            "".join(traceback.format_exception(error_type, error_message, traceback_))
        )
        console_writer.write_last_line(last_stack.lineno, line)
        console_writer.write_debug_tree(variables)
        console_writer.write_newline()

    if configs.has_notifier():
        notify_os(configs.title, configs.message)

def get_last_stack(traceback_: TracebackType):
    """Get the last stack and a new formated line of the last crashing line"""
    # Get last stack where crash occuredtrace
    last_stack = traceback.extract_tb(traceback_)[-1]

    # Get the source of the last stack
    source = Source(last_stack.filename, linecache.getlines(last_stack.filename))

    # Extract all relevant parts
    tokens = extrace_statement_from_source(source, last_stack)

    # Format into one line
    line = extract_source_code(tokens)

    return last_stack, line.strip()

def get_vars_from_tb(traceback_: TracebackType, line: str) -> List[Variable]:
    """Extract all variables from a line and a given traceback"""

    # Parse to get all names
    names = parse_error_line(line)

    # Retrieve locals and globals from the dead
    locals_, globals_ = retrieve_post_mortem_stack_infos(traceback_)

    # Get all variables and values
    variables = debug_variables(names, locals_, globals_)

    return variables

def extrace_statement_from_source(source: Source, last_stack) -> List[List[Token]]:
    """Get frame infos and get code pieces (from stack_data) by line
    of crash """
    pieces = source.pieces
    tokens = source.tokens_by_lineno

    statement_piece_tokens = []
    for piece in pieces:
        if last_stack.lineno in list(piece):
            for line in list(piece):
                statement_piece_tokens.append(tokens[line])

    return statement_piece_tokens

def extract_source_code(tokens: List[List[Token]]) -> str:
    """Get all strings of the tokens and flatten into on list"""
    statement_line = []
    for line_tokens in tokens:
        for token in line_tokens:
            string = token.string
            if not string.strip() == "":
                statement_line.append(string.strip())

    return " ".join(statement_line)

def debug_variables(variables: List[Variable], locals_: dict, globals_: dict) -> List[Variable]:
    """Evaluate for every given variable the value and type"""
    chainmap = ChainMap(locals_, globals_, vars(builtins))
    for var in variables:
        try:
            value = chainmap[var.name]
            var.value = value
        except KeyError:
            pass
    return variables

def parse_error_line(line: str) -> List[Variable]:
    """Parse a line of python code and extract all (variable) names from it"""
    variables = []
    try:
        # If we handling multilines this will not be parsed
        tree = ast.parse(line)

    except IndentationError as error:
        # Case of trying to parse a piece of a statement, like a for loop header
        # Try again with a pass stament
        print(line)
        tree = ast.parse(line + "pass")

    except SyntaxError as error:
        if error.args[0] == "unexpected EOF while parsing":
            try:
                tree = ast.parse(line + "pass")
            except SyntaxError as syntax_error:
                raise ParseError("Could not parse line: {}".format(line)) from syntax_error
        else:
            raise ParseError("Could not parse line: {}".format(line)) from error

    for node in ast.walk(tree):
        # For now just try to do it with names
        if isinstance(node, ast.Name):
            variables.append(Variable(node.id, node.col_offset))
    return variables

def format_line(line: str) -> str:
    """Try format a source code line"""

    formatted_line = line
    try:
        # If we handling multilines this will not be parsed
        formatted_line, _ =  FormatCode(line)

    except IndentationError as error:
        # Case of trying to parse a piece of a statement, like a for loop header
        # Try again with a pass stament
        formatted_line, _ = FormatCode(line + "pass")
        formatted_line = formatted_line.replace("pass", "")

    except SyntaxError as error:
        if error.args[0] == "unexpected EOF while parsing":
            try:
                formatted_line, _ = FormatCode(line + "pass")
                formatted_line = formatted_line.replace("pass", "")
            except SyntaxError as syntax_error:
                raise ParseError("Could not parse line: {}".format(line)) from syntax_error
        else:
            raise ParseError("Could not parse line: {}".format(line)) from error

    return formatted_line.strip()

def retrieve_post_mortem_stack_infos(traceback_: TracebackType):
    """Retrieve post mortem all local and global
    variables of given traceback"""

    base_debugger = Bdb()
    stack, i = base_debugger.get_stack(None, traceback_)

    # Get global and local vals
    locals_ = stack[i][0].f_locals
    globals_ = stack[i][0].f_globals
    return locals_, globals_



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['frosch']

package_data = \
{'': ['*']}

install_requires = \
['Pygments>=2.7.2,<3.0.0',
 'asttokens>=2.0.4,<3.0.0',
 'colorama>=0.4.4,<0.5.0',
 'stack-data>=0.1.0,<0.2.0',
 'yapf>=0.30.0,<0.31.0']

entry_points = \
{'console_scripts': ['example = example.example:hello']}

setup_kwargs = {
    'name': 'frosch',
    'version': '0.1.6',
    'description': 'Better runtime error messages',
    'long_description': '# frosch - Runtime Error Debugger\n\n[![PyPI version](https://badge.fury.io/py/frosch.svg)](https://badge.fury.io/py/frosch)\n![Codecov](https://img.shields.io/codecov/c/github/HallerPatrick/frosch)\n\nBetter runtime error messages\n\nAre you also constantly seeing the runtime error message the\npython interpreter is giving you?\nIt lacks some color and more debug information!\n\n\nGet some good looking error tracebacks and beautifuly formatted\nlast line with all its last values *before* you crashed the program.\n\n<h1 align="center" style="padding-left: 20px; padding-right: 20px">\n  <img src="resources/showcase.png">\n</h1>\n\n\n## Installation\n\n```bash\n$ pip install frosch\n```\n\n## Usage\n\nCall the hook function at the beginning of your program.\n\n```python\n\nfrom frosch import hook\n\nhook()\n\nx = 3 + "String"\n\n```\n\n### Print Exceptions\n\nYou can also easily print your catched exceptions to stdout\n\n```python\n\nfrom frosch import print_exception\n\ntry:\n  x = [0, 1]\n  x[3]\nexcept IndexError as error:\n  print_exception(error)\n\n```\n\n\n# Configuration\n\n## Themes\n\nfrosch allows to use different themes for styling the output:\n\n| Themes   |          |          |               |             |      |\n|----------|----------|----------|---------------|-------------|------|\n| abap     | bw       | igor     | native        | rrt         | trac |\n| algol    | colorful | inkpot   | paraiso_dark  | sas         | vim  |\n| algol_nu | default  | lovelace | paraiso_light | solarized   | vs   |\n| arduino  | emacs    | manni    | pastie        | stata_dark  | xcode |\n| autumn   | friendly | monokai  | perldoc       | stata_light |      |\n| borland  | fruity   | murphy   | rainbow_dash  | tango       |      |\n\nUsage:\n\n```python\nfrom frosch import hook\n\nhook(theme="vim")\n````\n\n## OS Notifications\n\nBut wait there is more!\n\nRunning longer scripts in the background?\n\nJust add a title and/or message to the hook and it will you give a notification when your program \nis crashing.\n\n\n```python\n\nfrom frosch import hook\n\nhook(\n  theme="vs", # VSCode Theme\n  title="I crashed!",\n  message="Run Number #1444 is also crashing..."\n)\n```\n\nThis works on MacOS (`osascript`), Linux (`notify-send`) and Windows (`powershell`). \n\n\n\n# Contribution\n\n`frosch` uses [poetry](https://github.com/python-poetry/poetry) for build and dependency\nmanagement, so please install beforehand.\n\n## Setup\n\n```bash\n$ git clone https://github.com/HallerPatrick/frosch.git\n$ poetry install\n```\n\n## Run tests\n\n```python\n$ python -m pytest tests\n```\n',
    'author': 'Patrick Haller',
    'author_email': 'patrickhaller40@googlemail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/HallerPatrick/frosch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
